package qrz

import (
	"fmt"
	"io/ioutil"
	"log"
	"net/http"
	"regexp"
	"strings"

	"git.paulbsd.com/paulbsd/qrz/src/config"
	"github.com/antchfx/htmlquery"
	"github.com/robfig/cron"
	"golang.org/x/text/encoding/charmap"
)

// InitCronConfig create task schedules
func InitCronConfig(config config.Config) (err error) {
	cr := cron.New()
	cr.AddFunc(config.Cron, func() { err = Run(config) })
	cr.Start()
	return
}

// Run runs the main loop
func Run(config config.Config) (err error) {
	for _, group := range config.QrzGroups {
		url := fmt.Sprintf(config.URLBase, group)
		fmt.Println(fmt.Sprintf("Processing the %s group", group))

		bodyStr, err := getBody(url)
		if err != nil {
			log.Fatalln(err)
		}

		frsPeople, err := getFrsEntries(config, bodyStr)
		if err != nil {
			log.Fatalln(err)
		}

		existingQrz, err := getCurrentEntries(config)
		if err != nil {
			log.Fatalln(err)
		}

		err = discardExistingEntries(config, &frsPeople, existingQrz)
		if err != nil {
			log.Fatalln(err)
		}

		if len(frsPeople) > 0 {
			err = insertFrsEntryToDB(config, frsPeople)
			if err != nil {
				log.Fatalln(err)
			}
		} else {
			fmt.Println(fmt.Sprintf("No entry to insert in database for group %s", group))
		}
	}
	return
}

// getBody fetch html page
func getBody(url string) (string, error) {
	var client http.Client

	resp, err := client.Get(url)
	if err != nil {
		return "", err
	}

	body, err := ioutil.ReadAll(resp.Body)
	if err != nil {
		return "", err
	}
	defer resp.Body.Close()

	dec := charmap.ISO8859_1.NewDecoder()
	output, _ := dec.Bytes(body)
	bodyString := string(output)

	return bodyString, nil
}

// getFrsEntries get FRS entries from html body
func getFrsEntries(config config.Config, body string) (frsentries map[string]FrsEntry, err error) {
	frsentries = make(map[string]FrsEntry)
	re := regexp.MustCompile(`^[0-9]{1,4}\s[A-Z]{1,4}\s[0-9]{1,4}`)

	htmlpage, err := htmlquery.Parse(strings.NewReader(body))
	if err != nil {
		return
	}

	for _, n := range htmlquery.Find(htmlpage, "//tr") {
		td := htmlquery.Find(n, "//td")
		if re.MatchString(htmlquery.InnerText(td[0])) {
			frs := FrsEntry{
				QRZ:     strings.Replace(htmlquery.InnerText(td[0]), "'", "\\'", -1),
				Name:    strings.Replace(htmlquery.InnerText(td[1]), "'", "\\'", -1),
				City:    strings.Replace(htmlquery.InnerText(td[2]), "'", "\\'", -1),
				Dept:    strings.Replace(htmlquery.InnerText(td[3]), "'", "\\'", -1),
				Country: strings.Replace(htmlquery.InnerText(td[4]), "'", "\\'", -1)}
			frsentries[frs.QRZ] = frs
		}
	}

	return
}

// getCurrentEntries fetch existing entries from database
func getCurrentEntries(config config.Config) (existingQRZ []string, err error) {
	q := fmt.Sprintf(config.Statements.DbGetQrz, config.DbTable)
	rows, err := config.Db.Query(q)

	for rows.Next() {
		var i string
		rows.Scan(&i)
		existingQRZ = append(existingQRZ, i)
	}

	return
}

// discardExistingEntries remove existing entries from original map[string]FrsEntry
func discardExistingEntries(config config.Config, frsPeople *map[string]FrsEntry, existingQRZ []string) (err error) {
	for _, entry := range existingQRZ {
		delete(*frsPeople, entry)
	}

	return
}

// insertFrsEntryToDB inserts frs entries to database
func insertFrsEntryToDB(config config.Config, frsPeople map[string]FrsEntry) (err error) {
	tx := config.Db.MustBegin()
	var qrzNum int

	fmt.Println(fmt.Sprintf("Starting inserts of %d entries", len(frsPeople)))

	for _, j := range frsPeople {
		query := fmt.Sprintf(config.Statements.DbInsert, config.DbTable, j.QRZ, j.Name, j.City, j.Dept, j.Country)
		tx.MustExec(query)
		qrzNum++
	}

	err = tx.Commit()
	if err != nil {
		return err
	}
	fmt.Println(fmt.Sprintf("Committed %d qrz entries", qrzNum))

	return nil
}

// FrsEntry describe FRS people
type FrsEntry struct {
	QRZ     string `db:"qrz"`
	Name    string `db:"name"`
	City    string `db:"city"`
	Dept    string `db:"dept"`
	Country string `db:"country"`
}
