package qrzws

import (
	"fmt"
	"log"
	"net/http"
	"strings"

	"git.paulbsd.com/paulbsd/qrz/src/config"
	"git.paulbsd.com/paulbsd/qrz/src/page"
	"git.paulbsd.com/paulbsd/qrz/src/qrz"
	"git.paulbsd.com/paulbsd/qrz/src/static"
	"git.paulbsd.com/paulbsd/qrz/src/templates"
	"github.com/gobuffalo/packr/v2"
	"github.com/labstack/echo/v4"
)

// RunServer runs the main echo server
func RunServer(config config.Config) (err error) {
	var p page.Page

	templatesbox := packr.New("templates", "./templates")
	staticbox := packr.New("static", "./static")

	builtTemplates, _ := templates.BuildTemplates(templatesbox)

	e := echo.New()
	e.HideBanner = true
	e.Renderer = builtTemplates

	e.GET("/", p.Index)
	e.GET("/static/*", func(c echo.Context) error {
		return static.GetStatic(staticbox, c)
	})
	e.POST("/qrzws", func(c echo.Context) (err error) {
		res, err := Run(c, config)
		if err != nil {
			log.Fatalln(err)
		}
		return c.JSON(http.StatusOK, res)
	})

	go qrz.Run(config)
	e.Logger.Fatal(e.Start(fmt.Sprintf(":%d", config.Port)))
	return
}

// Run runs the main loop
func Run(c echo.Context, config config.Config) (res QrzDatatableOutput, err error) {
	qrzinputjson := new(QrzDatatableInput)
	var count int

	if err = c.Bind(qrzinputjson); err != nil {
		return
	}
	query, err := BuildQuery(config, *qrzinputjson)

	rows, err := config.Db.Queryx(query)
	for rows.Next() {
		var line []string
		results := make(map[string]interface{})
		err = rows.MapScan(results)
		colslice, _ := rows.Columns()
		for _, column := range colslice {
			line = append(line, fmt.Sprintf("%s", results[column]))
		}
		res.Data = append(res.Data, line)
		count++
	}
	res.RecordsTotal = count
	res.RecordsFiltered = count
	res.Draw = qrzinputjson.Draw

	return
}

// BuildQuery builds query for SQL engine
func BuildQuery(config config.Config, qrzdt QrzDatatableInput) (query string, err error) {
	selectstatement, err := SetSelectStatement(config, qrzdt)
	orderstatement, err := SetOrderStatement(config, qrzdt)
	limitstatement, err := SetLimitStatement(config, qrzdt)
	searchstatement, err := SetSearchStatement(config, qrzdt)

	query = fmt.Sprintf("SELECT %s from %s %s %s %s;", selectstatement, config.DbTable, searchstatement, orderstatement, limitstatement)
	return
}

// SetSelectStatement build the sql select statement part
func SetSelectStatement(config config.Config, qrzdt QrzDatatableInput) (selectstatement string, err error) {
	var cols []string
	if len(qrzdt.Columns) > 0 {
		for _, col := range qrzdt.Columns {
			cols = append(cols, col.Name)
		}
		selectstatement = strings.Join(cols, ", ")
	} else {
		selectstatement = "*"
	}
	return
}

// SetOrderStatement build the sql order statement part
func SetOrderStatement(config config.Config, qrzdt QrzDatatableInput) (orderstmt string, err error) {
	var orderstmts []string
	for _, col := range qrzdt.Order {
		orderstmts = append(orderstmts, fmt.Sprintf("%s %s", qrzdt.Columns[col.Column].Name, strings.ToUpper(col.Dir)))
	}
	if len(orderstmts) > 0 {
		orderstmt = fmt.Sprintf("ORDER BY %s", strings.Join(orderstmts, ", "))
	}
	return
}

// SetLimitStatement build the sql limit statement part
func SetLimitStatement(config config.Config, qrzdt QrzDatatableInput) (limitstmt string, err error) {
	if qrzdt.Length < 1 {
		qrzdt.Length = 50
	}
	limitstmt = fmt.Sprintf("LIMIT %d OFFSET %d", qrzdt.Length, qrzdt.Start)
	return
}

// SetSearchStatement build the where clause in sql statement
func SetSearchStatement(config config.Config, qrzdt QrzDatatableInput) (searchstmt string, err error) {
	var searchstmtslice []string
	if len(qrzdt.Columns) > 0 {
		searchstmtslice = append(searchstmtslice, "WHERE")
		for id, i := range qrzdt.Columns {
			searchstmtslice = append(searchstmtslice, fmt.Sprintf("%s LIKE '%%%s%%'", i.Name, qrzdt.Search.Value))
			if id < len(qrzdt.Columns)-1 {
				searchstmtslice = append(searchstmtslice, "OR")
			}
		}
		searchstmt = strings.Join(searchstmtslice, " ")
	}
	return
}

// getCurrentEntries fetch existing entries from database
func getCurrentEntries(config config.Config) (existingQRZ []string, err error) {
	q := fmt.Sprintf(config.Statements.DbGetQrz, config.DbTable)
	rows, err := config.Db.Query(q)

	for rows.Next() {
		var qrzelem string
		rows.Scan(&qrzelem)
		existingQRZ = append(existingQRZ, qrzelem)
	}

	return
}

// QrzDatatableInput describes datatable query
type QrzDatatableInput struct {
	Columns []struct {
		Data       int    `json:"data"`
		Name       string `json:"name"`
		Searchable bool   `json:"searchable"`
		Orderable  bool   `json:"orderable"`
		Search     struct {
			Regex bool   `json:"regex"`
			Value string `json:"value"`
		} `json:"search"`
	} `json:"columns"`
	Draw   int `json:"draw"`
	Length int `json:"length"`
	Order  []struct {
		Column int    `json:"column"`
		Dir    string `json:"dir"`
	} `json:"order"`
	Search struct {
		Regex bool   `json:"regex"`
		Value string `json:"value"`
	} `json:"search"`
	Start int `json:"start"`
}

// QrzDatatableOutput describes datatable response
type QrzDatatableOutput struct {
	Data            [][]string `json:"data"`
	Draw            int        `json:"draw"`
	RecordsTotal    int        `json:"recordsTotal"`
	RecordsFiltered int        `json:"recordsFiltered"`
}
